# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Internal DB implementation for the cache.
"""

from __future__ import annotations

import logging
import os
import sqlite3
import sys
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path

log = logging.getLogger(__name__)

__all__ = [
    "CacheDB",
]

#
# Constants
#

_ERR_INVALID_PARAMETER_TYPE = (
    "Model parameters must be strings, integers, floats, or booleans, but got "
    "parameter {key}={value!r}"
)


#
# Classes
#


class CacheDB:
    """
    A model cache implemented as an SQLite database.
    """

    TBL_MODEL_PARAMS = "ModelParams"
    TBL_MODEL_RESPONSES = "ModelResponses"
    TBL_MODEL_CACHE = "ModelCache"

    def __init__(self, database: str | Path) -> None:
        """
        Initializes the CacheDB with a connection to the SQLite database.

        :param database: path to the SQLite database file; or ``:memory:`` to create an
            in-memory database
        """
        # Ensure the directory for the database exists
        if database != ":memory:":
            # Only attempt to create directory if not using an in-memory database
            directory = os.path.dirname(database)
            if directory and not os.path.exists(directory):
                log.warning(
                    f"Cache directory does not exist. Creating new directory: {directory}"
                )
                os.makedirs(directory)

        self.conn = sqlite3.connect(
            database=database,
            # We allow multiple threads to access the database connection; this is
            # necessary since a new thread may need to be created to start a new
            # asyncio event loop. We take care to ensure that DB transactions are
            # serialized by using the 'with self.conn' context manager.
            check_same_thread=False,
        )
        self._create_tables_and_indexes()

    def _create_tables_and_indexes(self) -> None:
        """
        Creates the necessary tables and indexes in the SQLite database.
        """
        with self.conn:
            self.conn.execute(
                """\
CREATE TABLE IF NOT EXISTS ModelCache (
    id INTEGER PRIMARY KEY,
    model_id TEXT NOT NULL,
    ctime DATETIME DEFAULT CURRENT_TIMESTAMP,
    atime DATETIME DEFAULT CURRENT_TIMESTAMP
);
"""
            )
            # We create a table for unique string values, to avoid redundancy
            # in the model parameters
            self.conn.execute(
                """\
CREATE TABLE IF NOT EXISTS UniqueStrings (
    id INTEGER PRIMARY KEY,
    value TEXT NOT NULL UNIQUE
);
"""
            )
            # Model parameters can bes stored as a reference to a unique string, or as a
            # integer, float, or boolean value
            self.conn.execute(
                """\
CREATE TABLE IF NOT EXISTS ModelParams (
    id INTEGER PRIMARY KEY,
    cache_id INTEGER,
    name TEXT,
    value_string_id INTEGER,
    value_int INTEGER,
    value_float REAL,
    FOREIGN KEY (cache_id) REFERENCES ModelCache(id)
);
"""
            )
            self.conn.execute(
                """\
CREATE TABLE IF NOT EXISTS ModelResponses (
    id INTEGER PRIMARY KEY,
    cache_id INTEGER,
    response TEXT NOT NULL,
    FOREIGN KEY (cache_id) REFERENCES ModelCache(id)
);
"""
            )
            self.conn.execute(
                """\
CREATE INDEX IF NOT EXISTS idx_prompt_model_id ON ModelCache(model_id);
"""
            )
            self.conn.execute(
                """\
CREATE INDEX IF NOT EXISTS idx_cache_id_name ON ModelParams(cache_id, name);
"""
            )

    def add_entry(
        self,
        *,
        model_id: str,
        responses: str | Iterable[str],
        **model_params: str | int | float | bool,
    ) -> None:
        """
        Add a new entry to the cache.

        :param model_id: the identifier of the model that generated the response
        :param responses: the response or responses generated by the model
        :param model_params: additional parameters used by the model
        :raises TypeError: if a model parameter is not a string, integer, float, or
            boolean
        """
        with self.conn:
            cursor = self.conn.execute(
                """\
INSERT INTO ModelCache (model_id)
VALUES (?);
""",
                (model_id,),
            )
            cache_id = cursor.lastrowid

            for response in [responses] if isinstance(responses, str) else responses:
                self.conn.execute(
                    """\
INSERT INTO ModelResponses (cache_id, response)
VALUES (?, ?);
""",
                    (cache_id, response),
                )

            if model_params:
                for key, value in model_params.items():
                    if isinstance(value, str):
                        cursor = self.conn.execute(
                            """\
INSERT OR IGNORE INTO UniqueStrings (value)
VALUES (?);
""",
                            (value,),
                        )
                        if cursor.rowcount == 0:
                            cursor = self.conn.execute(
                                """\
SELECT id FROM UniqueStrings WHERE value = ?;
""",
                                (value,),
                            )
                            value_string_id = cursor.fetchone()[0]
                        else:
                            value_string_id = cursor.lastrowid
                        self.conn.execute(
                            """\
INSERT INTO ModelParams (cache_id, name, value_string_id)
VALUES (?, ?, ?);
""",
                            (cache_id, key, value_string_id),
                        )
                    elif isinstance(value, (int, bool)):
                        self.conn.execute(
                            """\
INSERT INTO ModelParams (cache_id, name, value_int)
VALUES (?, ?, ?);
""",
                            (cache_id, key, value),
                        )
                    elif isinstance(value, float):
                        self.conn.execute(
                            """\
INSERT INTO ModelParams (cache_id, name, value_float)
VALUES (?, ?, ?);
""",
                            (cache_id, key, value),
                        )
                    else:
                        raise TypeError(
                            _ERR_INVALID_PARAMETER_TYPE.format(key=key, value=value)
                        )

    def get_entry(
        self,
        *,
        model_id: str,
        **model_params: str | int | float | bool,
    ) -> list[str] | None:
        """
        Retrieve an entry from the cache based on the prompt, model ID, and model
        parameters.

        :param model_id: the identifier of the model used
        :param model_params: additional used by the model
        :return: a list of responses generated by the model or ``None`` if no matching
            entry is found
        :raises TypeError: if a model parameter is not a string, integer, float, or
            boolean
        """
        base_query = """
    SELECT MC.id FROM ModelCache MC
    LEFT JOIN ModelParams MP ON MC.id = MP.cache_id
    WHERE MC.model_id = ?"""
        base_params = [model_id]

        param_query = ""
        param_params = []
        if model_params:
            subqueries = []
            for key, value in model_params.items():
                if isinstance(value, str):
                    subqueries.append(
                        "(MP.name = ? AND MP.value_string_id = "
                        "(SELECT id FROM UniqueStrings WHERE value = ?))"
                    )
                elif isinstance(value, (int, bool)):
                    subqueries.append("(MP.name = ? AND MP.value_int = ?)")
                elif isinstance(value, float):
                    subqueries.append("(MP.name = ? AND MP.value_float = ?)")
                else:
                    raise TypeError(
                        _ERR_INVALID_PARAMETER_TYPE.format(key=key, value=value)
                    )
                param_params.extend([key, value])
            param_query = f"""
    AND ({' OR '.join(subqueries)})"""

        final_query = f"""\
SELECT id
FROM ({base_query} {param_query}
    GROUP BY MC.id
    HAVING COUNT(DISTINCT MP.name) = ?
) AS FilteredCache
WHERE (SELECT COUNT(*) FROM ModelParams WHERE cache_id = FilteredCache.id) = ?
"""

        n_model_params = len(model_params) if model_params else 0
        final_params = [*base_params, *param_params, n_model_params, n_model_params]

        with self.conn:
            cursor = self.conn.execute(final_query, final_params)
            row = cursor.fetchone()

            if row:
                cache_id = row[0]

                response_cursor = self.conn.execute(
                    """\
SELECT response FROM ModelResponses WHERE cache_id = ?;
""",
                    (cache_id,),
                )
                responses = [
                    response_row[0] for response_row in response_cursor.fetchall()
                ]

                # update the last accessed time stamp
                self.conn.execute(
                    """\
UPDATE ModelCache
SET atime = CURRENT_TIMESTAMP
WHERE id = ?;
""",
                    (cache_id,),
                )

                return responses

        return None

    def count_entries(self) -> dict[str, int]:
        """
        Count the number of entries in the cache per model ID.

        :return: a dictionary with the number of entries for each model ID
        """
        with self.conn:
            cursor = self.conn.execute(
                """\
SELECT model_id, COUNT(*) FROM ModelCache GROUP BY model_id;
"""
            )
            return dict(cursor.fetchall())

    def get_earliest_creation_times(self) -> dict[str, datetime]:
        """
        Get the earliest creation time for each model ID in the cache.

        :return: a dictionary with the earliest creation time for each model ID
        """
        return self._get_times_per_model(field="ctime", func="MIN")

    def get_latest_creation_times(self) -> dict[str, datetime]:
        """
        Get the latest creation time for each model ID in the cache.

        :return: a dictionary with the latest creation time for each model ID
        """
        return self._get_times_per_model(field="ctime", func="MAX")

    def get_earliest_access_times(self) -> dict[str, datetime]:
        """
        Get the earliest access time for each model ID in the cache.

        :return: a dictionary with the earliest access time for each model ID
        """
        return self._get_times_per_model(field="atime", func="MIN")

    def get_latest_access_times(self) -> dict[str, datetime]:
        """
        Get the latest access time for each model ID in the cache.

        :return: a dictionary with the latest access time for each model ID
        """
        return self._get_times_per_model(field="atime", func="MAX")

    def clear(
        self,
        *,
        model_id: str | None = None,
        accessed_before: datetime | None = None,
        created_before: datetime | None = None,
        accessed_after: datetime | None = None,
        created_after: datetime | None = None,
    ) -> None:
        """
        Clears all entries from the cache, or only entries that match all specified
        conditions.

        :param model_id: delete only cache entries for this model ID (optional)
        :param accessed_before: delete only cache entries that were accessed before this
            iso_timestamp (optional)
        :param created_before: delete only cache entries that were created before this
            iso_timestamp (optional)
        :param accessed_after: delete only cache entries that were accessed after this
            iso_timestamp (optional)
        :param created_after: delete only cache entries that were created after this
            iso_timestamp (optional)
        """

        conn = self.conn

        def _as_UTC_string(dt: datetime) -> str:
            # convert to UTC and format as string
            return dt.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")

        with conn:
            conditions = []
            params = []
            if model_id:
                conditions.append("model_id = ?")
                params.append(model_id)
            if accessed_before:
                conditions.append("atime < ?")
                params.append(_as_UTC_string(accessed_before))
            if created_before:
                conditions.append("ctime < ?")
                params.append(_as_UTC_string(created_before))
            if accessed_after:
                conditions.append("atime > ?")
                params.append(_as_UTC_string(accessed_after))
            if created_after:
                conditions.append("ctime > ?")
                params.append(_as_UTC_string(created_after))
            if conditions:
                filter = " WHERE " + " AND ".join(conditions)
                filter_dependent = (
                    f" WHERE cache_id IN (SELECT id FROM ModelCache{filter})"
                )
            else:
                filter = filter_dependent = ""

            conn.execute("DELETE FROM ModelParams" + filter_dependent, params)
            conn.execute("DELETE FROM ModelResponses" + filter_dependent, params)
            conn.execute("DELETE FROM ModelCache" + filter, params)

            # Remove unused unique strings
            conn.execute(
                """\
DELETE FROM UniqueStrings
WHERE id NOT IN (SELECT DISTINCT value_string_id FROM ModelParams);
"""
            )

    def _get_times_per_model(self, *, field: str, func: str) -> dict[str, datetime]:
        with self.conn:
            cursor = self.conn.execute(
                f"""\
SELECT model_id, {func}({field}) FROM ModelCache GROUP BY model_id;
"""
            )
            return {
                model_id: _parse_iso_utc(timestamp) for model_id, timestamp in cursor
            }

    def __del__(self) -> None:
        """
        Closes the connection to the SQLite database.
        """
        self.conn.close()


#
# Auxiliary functions
#


def _parse_iso_utc(iso_timestamp: str) -> datetime:
    """
    Parse an ISO-formatted timestamp string and return a datetime object in UTC.

    If the timezone is not specified, the timestamp is assumed to be in UTC.

    If the timezone is specified, the timestamp will be converted to UTC.

    :param iso_timestamp: an ISO-formatted timestamp string
    :return: the corresponding datetime object in UTC
    """
    try:
        if sys.version_info >= (3, 11):
            return datetime.fromisoformat(iso_timestamp + "Z")
        else:
            # The "Z" suffix is not supported in Python versions before 3.11
            return datetime.fromisoformat(iso_timestamp + "+00:00")
    except ValueError:
        # Try to parse the iso_timestamp without the 'Z' suffix
        return datetime.fromisoformat(iso_timestamp).astimezone(timezone.utc)



import os
from collections.abc import Generator
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
from httpx import HTTPStatusError, Response

from artkit.model.llm.base import HTTPXChatConnector
from artkit.model.llm.history import ChatHistory
from artkit.model.util import RateLimitException

#######################################################################################
#                                     Constants                                       #
#######################################################################################
MODEL_ID = "test_model"
URL = "http://test.url"

MESSAGE = "What is the color of the sky?"
RESPONSE_TEXT = "The sky is blue."


@pytest.mark.asyncio
async def test_get_response(mock_custom_connector: HTTPXChatConnector) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        mock_post = Mock()
        mock_post.json = Mock(return_value={"results": [{"outputText": RESPONSE_TEXT}]})
        mock_post.text = AsyncMock()
        mock_post.return_value.status_code = 200

        mock_connection = AsyncMock()
        mock_connection.request.return_value = mock_post
        MockClientSession.return_value = mock_connection

        response = await mock_custom_connector.get_response(message=MESSAGE)
        assert response == [RESPONSE_TEXT]


@pytest.mark.asyncio
async def test_rate_limit_error(
    mock_custom_connector: HTTPXChatConnector,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_rate_limit_error() -> None:
            request = Mock()
            response = Response(
                status_code=429,
                content=b"Rate limit exceeded",
                request=request,
            )
            err = HTTPStatusError(
                response=response,
                message="Rate limit exceeded",
                request=request,
            )
            raise err

        mock_connection.request.return_value.raise_for_status = raise_rate_limit_error
        MockClientSession.return_value = mock_connection

        with pytest.raises(RateLimitException):
            await mock_custom_connector.get_response(message=MESSAGE)


@pytest.mark.asyncio
async def test_invalid_request_error(
    mock_custom_connector: HTTPXChatConnector,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_invalid_request_error() -> None:
            request = Mock()
            response = Response(
                status_code=422,
                content=b"Internal server error",
                request=request,
            )
            err = HTTPStatusError(
                response=response,
                message="Invalid request",
                request=request,
            )
            raise err

        mock_connection.request.return_value.raise_for_status = (
            raise_invalid_request_error
        )
        MockClientSession.return_value = mock_connection

        with pytest.raises(ValueError):
            await mock_custom_connector.get_response(message=MESSAGE)


@pytest.mark.asyncio
async def test_unexpected_error(
    mock_custom_connector: HTTPXChatConnector,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_unexpected_error() -> None:
            request = Mock()
            response = Response(
                status_code=500,
                content=b"Internal server error",
                request=request,
            )
            err = HTTPStatusError(
                response=response,
                message="Internal server error",
                request=request,
            )
            raise err

        mock_connection.request.return_value.raise_for_status = raise_unexpected_error
        MockClientSession.return_value = mock_connection

        with pytest.raises(HTTPStatusError):
            await mock_custom_connector.get_response(message=MESSAGE)


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################
@pytest.fixture(scope="function")
def mock_custom_connector() -> Generator[HTTPXChatConnector, None, None]:
    class MockCustomChatEndpointConnector(HTTPXChatConnector):
        @classmethod
        def get_default_api_key_env(cls) -> str:
            """[see superclass]"""
            return "TEST"

        def build_request_arguments(
            self,
            message: str,
            *,
            history: ChatHistory | None = None,
            **model_params: dict[str, Any],
        ) -> dict[str, Any]:
            return dict(
                method="POST",
                url=self.model_id,
                data={"message": f"Formatted: {message}"},
                headers={"Authorization": "Bearer test_token"},
            )

        def parse_httpx_response(self, response: Response) -> list[str]:
            json_response = response.json()
            return [result["outputText"] for result in json_response["results"]]

    api_key_env = "TEST"
    os.environ[api_key_env] = "test"

    yield MockCustomChatEndpointConnector(
        model_id=URL,
        api_key_env=api_key_env,
        initial_delay=0.1,
        exponential_base=1.5,
        max_retries=2,
    )



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

import asyncio
import logging
import random
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import Any, Concatenate, ParamSpec, TypeVar

# from ..llm.base import ChatModel
from ...util import LogThrottlingHandler
from ..base import ConnectorMixin
from ._exception import RateLimitException

__all__ = ["retry_with_exponential_backoff", "retry_function_with_exponential_backoff"]


log = logging.getLogger(__name__)
# 5-second interval, max 5 messages
_throttling_handler = LogThrottlingHandler(
    handler=logging.StreamHandler(), interval=5, max_messages=5
)
_throttling_handler.setLevel(logging.WARNING)
log.addHandler(_throttling_handler)

#
# Type variables
#
# Naming convention used here:
# _ret for covariant type variables used in return positions
# _arg for contravariant type variables used in argument positions
#

T_ret = TypeVar("T_ret", covariant=True)
T_Connector = TypeVar("T_Connector", bound=ConnectorMixin[Any])
T_ChatModel = TypeVar("T_ChatModel")
P = ParamSpec("P")


def retry_function_with_exponential_backoff(
    func: Callable[Concatenate[T_ChatModel, P], Coroutine[Any, Any, T_ret]],
    max_retries: int,
    delay: float,
    exponential_base: float,
    jitter: bool,
) -> Callable[Concatenate[T_ChatModel, P], Coroutine[Any, Any, T_ret]]:
    """
    Decorator that adds retrying with exponential backoff to a ChatModel method.

    :param func: the function to decorate
    :param max_retries: the maximum number of retries
    :param delay: the delay between retries
    :param exponential_base: the base exponential factor
    :param jitter: the jittering factor

    :return: the decorated function
    """

    @wraps(func)
    async def exponential_delay(
        self: T_ChatModel, /, *args: P.args, **kwargs: P.kwargs
    ) -> T_ret:
        initial_delay = delay
        for num_retries in range(1, max_retries + 1):
            try:
                return await func(self, *args, **kwargs)

            # Retry on rate limit errors
            except RateLimitException as e:
                # Increment the delay
                initial_delay *= exponential_base * (1 + jitter * random.random())

                # Log retry
                log.warning(
                    "Rate limit exceeded: Retrying after model I/O error "
                    "(%s/%s). Wait for %.2f seconds.",
                    num_retries,
                    max_retries,
                    initial_delay,
                )

                # Sleep for the delay
                await asyncio.sleep(initial_delay)
                last_exception = e

        raise RateLimitException(
            "Rate limit error after max retries."
        ) from last_exception

    return exponential_delay


def retry_with_exponential_backoff(
    func: Callable[Concatenate[T_Connector, P], Coroutine[Any, Any, T_ret]]
) -> Callable[Concatenate[T_Connector, P], Coroutine[Any, Any, T_ret]]:
    """
    Decorator that adds retrying with exponential backoff to a connector method.

    It uses the retry_function_with_exponential_backoff decorator to decorate.

    :param func: the function to decorate
    :return: the decorated function
    """

    @wraps(func)
    async def _wrapper(
        self: T_Connector, /, *args: P.args, **kwargs: P.kwargs
    ) -> T_ret:
        """
        Wrapper function for retrying a function with exponential backoff.

        :param self: the connector instance the method is called on
        :param args: the positional arguments passed to the function
        :param kwargs: the keyword arguments passed to the function

        :return: the result of the function
        """
        # Initialize variables
        delay = self.initial_delay

        # Loop until a successful response or max_retries is hit
        # or an uncaught exception is raised
        max_retries = self.max_retries
        exponential_base = self.exponential_base
        jitter = self.jitter

        return await retry_function_with_exponential_backoff(
            func=func,
            max_retries=max_retries,
            delay=delay,
            exponential_base=exponential_base,
            jitter=jitter,
        )(self, *args, **kwargs)

    if func.__code__ is _wrapper.__code__:  # type: ignore[attr-defined]
        # If the function is already wrapped with our decorator, return it as is.
        # This can happen if a connector's __init_subclass__ method is called
        # multiple times along the MRO.
        return func
    else:
        return _wrapper



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the Groq llm module.
"""

import logging
from collections.abc import Iterator
from contextlib import AsyncExitStack
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, inheritdoc

from ...util import RateLimitException
from ..base import ChatModelConnector
from ..history import ChatHistory

log = logging.getLogger(__name__)

try:  # pragma: no cover
    from groq import AsyncGroq, RateLimitError
    from groq.types.chat import ChatCompletion
except ImportError:

    class AsyncGroq(metaclass=MissingClassMeta, module="groq"):  # type: ignore
        """
        Placeholder class for missing ``AsyncGroq`` class.
        """

    class ChatCompletion(  # type: ignore
        metaclass=MissingClassMeta, module="groq.types.chat"
    ):
        """Placeholder class for missing ``ChatCompletion`` class."""

    class RateLimitError(metaclass=MissingClassMeta, module="groq"):  # type: ignore
        """Placeholder class for missing ``RateLimitError`` class."""


__all__ = [
    "GroqChat",
]


#
# Type variables
#

T_GroqLLM = TypeVar("T_GroqLLM", bound="GroqChat")

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class GroqChat(ChatModelConnector[AsyncGroq]):
    """
    An asynchronous Groq LLM.
    """

    @classmethod
    def get_default_api_key_env(cls) -> str:
        """[see superclass]"""
        return "GROQ_API_KEY"

    async def get_response(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> list[str]:
        """[see superclass]"""
        async with AsyncExitStack():

            try:
                completion = await self.get_client().chat.completions.create(
                    messages=list(
                        self._messages_to_groq_format(message, history=history)
                    ),
                    model=self.model_id,
                    **{**self.get_model_params(), **model_params},
                )
            except RateLimitError as e:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e

        return list(self._responses_from_completion(completion))

    def _make_client(self) -> AsyncGroq:  # pragma: no cover
        """[see superclass]"""
        return AsyncGroq(api_key=self.get_api_key())

    def _messages_to_groq_format(
        self, user_message: str, *, history: ChatHistory | None = None
    ) -> Iterator[dict[str, str]]:
        """
        Get the messages to send to the Groq LLM, based on the given user prompt
        and chat history, and the system prompt for this LLM.

        :param user_message: the user prompt to send to the Groq LLM
        :param history: the chat history to include in the messages (optional)
        :return: the messages object, in the format expected by the Groq API
        """
        if self.system_prompt:
            yield {"role": "system", "content": self.system_prompt}

        if history is not None:
            for message in history.messages:
                yield {"role": message.role, "content": message.text}

        yield {"role": "user", "content": user_message}

    @staticmethod
    def _responses_from_completion(
        completion: ChatCompletion,
    ) -> Iterator[str]:
        """
        Get the response from the given chat completion.

        :param completion: the chat completion to process
        :return: the alternate responses from the chat completion
        """

        for choice in completion.choices:
            message = choice.message
            if message.role != "assistant":
                log.warning(
                    "Expected only assistant messages, but got completion choice "
                    f"{choice!r}"
                )
            yield str(message.content)
